"""
Git Fileserver Backend

With this backend, branches and tags in a remote git repository are exposed to
salt as different environments.

To enable, add ``gitfs`` to the :conf_master:`fileserver_backend` option in the
Master config file.

.. code-block:: yaml

    fileserver_backend:
      - gitfs

.. note::
    ``git`` also works here. Prior to the 2018.3.0 release, *only* ``git``
    would work.

The Git fileserver backend supports both pygit2_ and GitPython_, to provide the
Python interface to git. If both are present, the order of preference for which
one will be chosen is the same as the order in which they were listed: pygit2,
then GitPython.

An optional master config parameter (:conf_master:`gitfs_provider`) can be used
to specify which provider should be used, in the event that compatible versions
of both pygit2_ and GitPython_ are installed.

More detailed information on how to use GitFS can be found in the :ref:`GitFS
Walkthrough <tutorial-gitfs>`.

.. note:: Minimum requirements

    To use pygit2_ for GitFS requires a minimum pygit2_ version of 0.20.3.
    pygit2_ 0.20.3 requires libgit2_ 0.20.0. pygit2_ and libgit2_ are developed
    alongside one another, so it is recommended to keep them both at the same
    major release to avoid unexpected behavior. For example, pygit2_ 0.21.x
    requires libgit2_ 0.21.x, pygit2_ 0.22.x will require libgit2_ 0.22.x, etc.

    To use GitPython_ for GitFS requires a minimum GitPython version of 0.3.0,
    as well as the git CLI utility. Instructions for installing GitPython can
    be found :ref:`here <gitfs-dependencies>`.

    To clear stale refs the git CLI utility must also be installed.

.. _pygit2: https://github.com/libgit2/pygit2
.. _libgit2: https://libgit2.github.com/
.. _GitPython: https://github.com/gitpython-developers/GitPython
"""


import logging

import salt.utils.gitfs
from salt.exceptions import FileserverConfigError

PER_REMOTE_OVERRIDES = (
    "base",
    "fallback",
    "mountpoint",
    "root",
    "ssl_verify",
    "saltenv_whitelist",
    "saltenv_blacklist",
    "refspecs",
    "disable_saltenv_mapping",
    "ref_types",
    "update_interval",
)
PER_REMOTE_ONLY = ("all_saltenvs", "name", "saltenv")

# Auth support (auth params can be global or per-remote, too)
AUTH_PROVIDERS = ("pygit2",)
AUTH_PARAMS = ("user", "password", "pubkey", "privkey", "passphrase", "insecure_auth")


log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "gitfs"
__virtual_aliases__ = ("git",)


def _gitfs(init_remotes=True):
    return salt.utils.gitfs.GitFS(
        __opts__,
        __opts__["gitfs_remotes"],
        per_remote_overrides=PER_REMOTE_OVERRIDES,
        per_remote_only=PER_REMOTE_ONLY,
        init_remotes=init_remotes,
    )


def __virtual__():
    """
    Only load if the desired provider module is present and gitfs is enabled
    properly in the master config file.
    """
    if __virtualname__ not in __opts__["fileserver_backend"]:
        return False
    try:
        _gitfs(init_remotes=False)
        # Initialization of the GitFS object did not fail, so we know we have
        # valid configuration syntax and that a valid provider was detected.
        return __virtualname__
    except FileserverConfigError:
        pass
    return False


def clear_cache():
    """
    Completely clear gitfs cache
    """
    return _gitfs(init_remotes=False).clear_cache()


def clear_lock(remote=None, lock_type="update"):
    """
    Clear update.lk
    """
    return _gitfs().clear_lock(remote=remote, lock_type=lock_type)


def lock(remote=None):
    """
    Place an update.lk

    ``remote`` can either be a dictionary containing repo configuration
    information, or a pattern. If the latter, then remotes for which the URL
    matches the pattern will be locked.
    """
    return _gitfs().lock(remote=remote)


def update(remotes=None):
    """
    Execute a git fetch on all of the repos
    """
    _gitfs().update(remotes)


def update_intervals():
    """
    Returns the update intervals for each configured remote
    """
    return _gitfs().update_intervals()


def envs(ignore_cache=False):
    """
    Return a list of refs that can be used as environments
    """
    return _gitfs().envs(ignore_cache=ignore_cache)


def find_file(path, tgt_env="base", **kwargs):  # pylint: disable=W0613
    """
    Find the first file to match the path and ref, read the file out of git
    and send the path to the newly cached file
    """
    return _gitfs().find_file(path, tgt_env=tgt_env, **kwargs)


def init():
    """
    Initialize remotes. This is only used by the master's pre-flight checks,
    and is not invoked by GitFS.
    """
    _gitfs()


def serve_file(load, fnd):
    """
    Return a chunk from a file based on the data received
    """
    return _gitfs().serve_file(load, fnd)


def file_hash(load, fnd):
    """
    Return a file hash, the hash type is set in the master config file
    """
    return _gitfs().file_hash(load, fnd)


def file_list(load):
    """
    Return a list of all files on the file server in a specified
    environment (specified as a key within the load dict).
    """
    return _gitfs().file_list(load)


def file_list_emptydirs(load):  # pylint: disable=W0613
    """
    Return a list of all empty directories on the master
    """
    # Cannot have empty dirs in git
    return []


def dir_list(load):
    """
    Return a list of all directories on the master
    """
    return _gitfs().dir_list(load)


def symlink_list(load):
    """
    Return a dict of all symlinks based on a given path in the repo
    """
    return _gitfs().symlink_list(load)



"""
Utility functions to modify other functions
"""


import logging
import types

import salt.utils.args
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)


def namespaced_function(function, global_dict, defaults=None, preserve_context=False):
    """
    Redefine (clone) a function under a different globals() namespace scope

        preserve_context:
            Allow keeping the context taken from orignal namespace,
            and extend it with globals() taken from
            new targetted namespace.
    """
    if defaults is None:
        defaults = function.__defaults__

    if preserve_context:
        _global_dict = function.__globals__.copy()
        _global_dict.update(global_dict)
        global_dict = _global_dict
    new_namespaced_function = types.FunctionType(
        function.__code__,
        global_dict,
        name=function.__name__,
        argdefs=defaults,
        closure=function.__closure__,
    )
    new_namespaced_function.__dict__.update(function.__dict__)
    return new_namespaced_function


def alias_function(fun, name, doc=None):
    """
    Copy a function
    """
    alias_fun = types.FunctionType(
        fun.__code__,
        fun.__globals__,
        str(name),
        fun.__defaults__,
        fun.__closure__,
    )
    alias_fun.__dict__.update(fun.__dict__)

    if doc and isinstance(doc, str):
        alias_fun.__doc__ = doc
    else:
        orig_name = fun.__name__
        alias_msg = "\nThis function is an alias of ``{}``.\n".format(orig_name)
        alias_fun.__doc__ = alias_msg + (fun.__doc__ or "")

    return alias_fun


def parse_function(function_arguments):
    """
    Helper function to parse function_arguments (module.run format)
    into args and kwargs.
    This function is similar to salt.utils.data.repack_dictlist, except that this
    handles mixed (i.e. dict and non-dict) arguments in the input list.

    :param list function_arguments: List of items and dicts with kwargs.

    :rtype: dict
    :return: Dictionary with ``args`` and ``kwargs`` keyword.
    """
    function_args = []
    function_kwargs = {}
    for item in function_arguments:
        if isinstance(item, dict):
            function_kwargs.update(item)
        else:
            function_args.append(item)
    return {"args": function_args, "kwargs": function_kwargs}


def call_function(salt_function, *args, **kwargs):
    """
    Calls a function from the specified module.

    :param function salt_function: Function reference to call
    :return: The result of the function call
    """
    argspec = salt.utils.args.get_function_argspec(salt_function)
    # function_kwargs is initialized to a dictionary of keyword arguments the function to be run accepts
    function_kwargs = dict(
        zip(
            argspec.args[
                -len(argspec.defaults or []) :
            ],  # pylint: disable=incompatible-py3-code
            argspec.defaults or [],
        )
    )
    # expected_args is initialized to a list of positional arguments that the function to be run accepts
    expected_args = argspec.args[
        : len(argspec.args or []) - len(argspec.defaults or [])
    ]
    function_args, kw_to_arg_type = [], {}
    for funcset in reversed(args or []):
        if not isinstance(funcset, dict):
            # We are just receiving a list of args to the function to be run, so just append
            # those to the arg list that we will pass to the func.
            function_args.append(funcset)
        else:
            for kwarg_key in funcset.keys():
                # We are going to pass in a keyword argument. The trick here is to make certain
                # that if we find that in the *args* list that we pass it there and not as a kwarg
                if kwarg_key in expected_args:
                    kw_to_arg_type[kwarg_key] = funcset[kwarg_key]
                else:
                    # Otherwise, we're good and just go ahead and pass the keyword/value pair into
                    # the kwargs list to be run.
                    function_kwargs.update(funcset)
    function_args.reverse()
    # Add kwargs passed as kwargs :)
    function_kwargs.update(kwargs)
    for arg in expected_args:
        if arg in kw_to_arg_type:
            function_args.append(kw_to_arg_type[arg])
    _exp_prm = len(argspec.args or []) - len(argspec.defaults or [])
    _passed_prm = len(function_args)
    missing = []
    if _exp_prm > _passed_prm:
        for arg in argspec.args[_passed_prm:]:
            if arg not in function_kwargs:
                missing.append(arg)
            else:
                # Found the expected argument as a keyword
                # increase the _passed_prm count
                _passed_prm += 1
    if missing:
        raise SaltInvocationError("Missing arguments: {}".format(", ".join(missing)))
    elif _exp_prm > _passed_prm:
        raise SaltInvocationError(
            "Function expects {} positional parameters, got only {}".format(
                _exp_prm, _passed_prm
            )
        )

    return salt_function(*function_args, **function_kwargs)



"""
    :codeauthor: :email:`Shane Lee <slee@saltstack.com>`
"""

import salt.grains.fibre_channel as fibre_channel
from tests.support.mock import MagicMock, mock_open, patch
from tests.support.unit import TestCase


class FibreChannelGrainsTestCase(TestCase):
    """
    Test cases for iscsi grains
    """

    def test_windows_fibre_channel_wwns_grains(self):
        wwns = [
            "20:00:00:25:b5:11:11:4c",
            "20:00:00:25:b5:11:11:5c",
            "20:00:00:25:b5:44:44:4c",
            "20:00:00:25:b5:44:44:5c",
        ]
        cmd_run_mock = MagicMock(return_value=wwns)
        with patch("salt.modules.cmdmod.powershell", cmd_run_mock):
            ret = fibre_channel._windows_wwns()
        assert ret == wwns, ret

    def test_linux_fibre_channel_wwns_grains(self):

        contents = ["0x500143802426baf4", "0x500143802426baf5"]
        files = ["file1", "file2"]
        with patch("glob.glob", MagicMock(return_value=files)), patch(
            "salt.utils.files.fopen", mock_open(read_data=contents)
        ):
            ret = fibre_channel._linux_wwns()

        assert ret == ["500143802426baf4", "500143802426baf5"], ret



import io

import pytest
import salt.modules.nilrt_ip as nilrt_ip
from tests.support.mock import patch


@pytest.fixture
def configure_loader_modules():
    return {nilrt_ip: {}}


@pytest.fixture
def patched_config_file():
    config_file = io.StringIO(
        """
    [some_section]
    name = thing
    fnord = bar
    icanhazquotes = "this string is quoted"
    icannothazquotes = this string is unquoted
    number_value = 42
    """
    )
    with patch("salt.utils.files.fopen", return_value=config_file):
        yield


def test_when_config_has_quotes_around_string_they_should_be_removed(
    patched_config_file,
):
    expected_value = "this string is quoted"
    option = "icanhazquotes"

    actual_value = nilrt_ip._load_config("some_section", [option])[option]

    assert actual_value == expected_value


def test_when_config_has_no_quotes_around_string_it_should_be_returned_as_is(
    patched_config_file,
):
    expected_value = "this string is unquoted"
    option = "icannothazquotes"

    actual_value = nilrt_ip._load_config("some_section", [option])[option]

    assert actual_value == expected_value


@pytest.mark.parametrize(
    "default_value",
    [
        42,
        -99.9,
        ('"', "some value", 42, '"'),
        ['"', "a weird list of values", '"'],
        {"this": "dictionary", "has": "multiple values", 0: '"', -1: '"'},
    ],
)
def test_when_default_value_is_not_a_string_and_option_is_missing_the_default_value_should_be_returned(
    patched_config_file, default_value
):
    option = "non existent option"

    actual_value = nilrt_ip._load_config(
        "some_section", options=[option], default_value=default_value
    )[option]

    assert actual_value == default_value



"""
Watch files and translate the changes into salt events.

.. versionadded:: 2019.2.0

:depends:   - watchdog Python module >= 0.8.3

"""

import collections
import logging

try:
    from watchdog.observers import Observer
    from watchdog.events import FileSystemEventHandler

    HAS_WATCHDOG = True
except ImportError:
    HAS_WATCHDOG = False

    class FileSystemEventHandler:
        """A dummy class to make the import work"""

        def __init__(self):
            pass


__virtualname__ = "watchdog"

log = logging.getLogger(__name__)

DEFAULT_MASK = [
    "create",
    "delete",
    "modify",
    "move",
]


class Handler(FileSystemEventHandler):
    def __init__(self, queue, masks=None):
        super().__init__()
        self.masks = masks or DEFAULT_MASK
        self.queue = queue

    def on_created(self, event):
        self._append_if_mask(event, "create")

    def on_modified(self, event):
        self._append_if_mask(event, "modify")

    def on_deleted(self, event):
        self._append_if_mask(event, "delete")

    def on_moved(self, event):
        self._append_if_mask(event, "move")

    def _append_if_mask(self, event, mask):
        logging.debug(event)

        self._append_path_if_mask(event, mask)

    def _append_path_if_mask(self, event, mask):
        if mask in self.masks:
            self.queue.append(event)


def __virtual__():
    if HAS_WATCHDOG:
        return __virtualname__
    return False


def _get_queue(config):
    """
    Check the context for the notifier and construct it if not present
    """

    if "watchdog.observer" not in __context__:
        queue = collections.deque()
        observer = Observer()
        for path in config.get("directories", {}):
            path_params = config.get("directories").get(path)
            masks = path_params.get("mask", DEFAULT_MASK)
            event_handler = Handler(queue, masks)
            observer.schedule(event_handler, path)

        observer.start()

        __context__["watchdog.observer"] = observer
        __context__["watchdog.queue"] = queue

    return __context__["watchdog.queue"]


class ValidationError(Exception):
    pass


def validate(config):
    """
    Validate the beacon configuration
    """

    try:
        _validate(config)
        return True, "Valid beacon configuration"
    except ValidationError as error:
        return False, str(error)


def _validate(config):
    if not isinstance(config, list):
        raise ValidationError("Configuration for watchdog beacon must be a list.")

    _config = {}
    for part in config:
        _config.update(part)

    if "directories" not in _config:
        raise ValidationError(
            "Configuration for watchdog beacon must include directories."
        )

    if not isinstance(_config["directories"], dict):
        raise ValidationError(
            "Configuration for watchdog beacon directories must be a dictionary."
        )

    for path in _config["directories"]:
        _validate_path(_config["directories"][path])


def _validate_path(path_config):
    if not isinstance(path_config, dict):
        raise ValidationError(
            "Configuration for watchdog beacon directory path must be a dictionary."
        )

    if "mask" in path_config:
        _validate_mask(path_config["mask"])


def _validate_mask(mask_config):
    valid_mask = [
        "create",
        "modify",
        "delete",
        "move",
    ]

    if not isinstance(mask_config, list):
        raise ValidationError("Configuration for watchdog beacon mask must be list.")

    if any(mask not in valid_mask for mask in mask_config):
        raise ValidationError("Configuration for watchdog beacon contains invalid mask")


def to_salt_event(event):
    return {
        "tag": __virtualname__,
        "path": event.src_path,
        "change": event.event_type,
    }


def beacon(config):
    """
    Watch the configured directories

    Example Config

    .. code-block:: yaml

        beacons:
          watchdog:
            - directories:
                /path/to/dir:
                  mask:
                    - create
                    - modify
                    - delete
                    - move

    The mask list can contain the following events (the default mask is create,
    modify delete, and move):
    * create  - File or directory is created in watched directory
    * modify  - The watched directory is modified
    * delete  - File or directory is deleted from watched directory
    * move    - File or directory is moved or renamed in the watched directory
    """

    _config = {}
    list(map(_config.update, config))

    queue = _get_queue(_config)

    ret = []
    while queue:
        ret.append(to_salt_event(queue.popleft()))

    return ret


def close(config):
    observer = __context__.pop("watchdog.observer", None)

    if observer:
        observer.stop()



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import salt.modules.pyenv as pyenv
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class PyenvTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.pyenv
    """

    def setup_loader_modules(self):
        return {pyenv: {}}

    # 'install' function tests: 1

    def test_install(self):
        """
        Test if it install pyenv systemwide
        """
        mock_opt = MagicMock(return_value="salt stack")
        mock_ret = MagicMock(return_value=0)
        with patch.dict(
            pyenv.__salt__, {"config.option": mock_opt, "cmd.retcode": mock_ret}
        ):
            self.assertTrue(pyenv.install())

    # 'update' function tests: 1

    def test_update(self):
        """
        Test if it updates the current versions of pyenv and python-Build
        """
        mock_opt = MagicMock(return_value="salt stack")
        with patch.dict(pyenv.__salt__, {"config.option": mock_opt}):
            self.assertFalse(pyenv.update())

    # 'is_installed' function tests: 1

    def test_is_installed(self):
        """
        Test if it check if pyenv is installed.
        """
        mock_cmd = MagicMock(return_value=True)
        mock_opt = MagicMock(return_value="salt stack")
        with patch.dict(
            pyenv.__salt__, {"config.option": mock_opt, "cmd.has_exec": mock_cmd}
        ):
            self.assertTrue(pyenv.is_installed())

    # 'install_python' function tests: 1

    def test_install_python(self):
        """
        Test if it install a python implementation.
        """
        mock_opt = MagicMock(return_value="salt stack")
        mock_cmd = MagicMock(return_value=True)
        mock_all = MagicMock(
            return_value={"retcode": 0, "stdout": "salt", "stderr": "error"}
        )
        with patch.dict(pyenv.__grains__, {"os": "Linux"}):
            mock_all = MagicMock(
                return_value={"retcode": 0, "stdout": "salt", "stderr": "error"}
            )
            with patch.dict(
                pyenv.__salt__,
                {
                    "config.option": mock_opt,
                    "cmd.has_exec": mock_cmd,
                    "cmd.run_all": mock_all,
                },
            ):
                self.assertEqual(pyenv.install_python("2.0.0-p0"), "error")

            mock_all = MagicMock(
                return_value={"retcode": True, "stdout": "salt", "stderr": "error"}
            )
            with patch.dict(
                pyenv.__salt__,
                {
                    "config.option": mock_opt,
                    "cmd.has_exec": mock_cmd,
                    "cmd.run_all": mock_all,
                },
            ):
                self.assertFalse(pyenv.install_python("2.0.0-p0"))

    # 'uninstall_python' function tests: 1

    def test_uninstall_python(self):
        """
        Test if it uninstall a python implementation.
        """
        mock_opt = MagicMock(return_value="salt stack")
        mock_cmd = MagicMock(return_value=True)
        mock_all = MagicMock(
            return_value={"retcode": True, "stdout": "salt", "stderr": "error"}
        )
        with patch.dict(
            pyenv.__salt__,
            {
                "cmd.has_exec": mock_cmd,
                "config.option": mock_opt,
                "cmd.run_all": mock_all,
            },
        ):
            self.assertTrue(pyenv.uninstall_python("2.0.0-p0"))

    # 'versions' function tests: 1

    def test_versions(self):
        """
        Test if it list the installed versions of python.
        """
        mock_opt = MagicMock(return_value="salt stack")
        mock_cmd = MagicMock(return_value=True)
        mock_all = MagicMock(
            return_value={"retcode": True, "stdout": "salt", "stderr": "error"}
        )
        with patch.dict(
            pyenv.__salt__,
            {
                "cmd.has_exec": mock_cmd,
                "config.option": mock_opt,
                "cmd.run_all": mock_all,
            },
        ):
            self.assertListEqual(pyenv.versions(), [])

    # 'default' function tests: 1

    def test_default(self):
        """
        Test if it returns or sets the currently defined default python.
        """
        mock_opt = MagicMock(return_value="salt stack")
        mock_cmd = MagicMock(return_value=True)
        mock_all = MagicMock(
            return_value={"retcode": True, "stdout": "salt", "stderr": "error"}
        )
        with patch.dict(
            pyenv.__salt__,
            {
                "cmd.has_exec": mock_cmd,
                "config.option": mock_opt,
                "cmd.run_all": mock_all,
            },
        ):
            self.assertEqual(pyenv.default(), "")
            self.assertTrue(pyenv.default("2.0.0-p0"))

    # 'list_' function tests: 1

    def test_list(self):
        """
        Test if it list the installable versions of python.
        """
        mock_opt = MagicMock(return_value="salt stack")
        mock_cmd = MagicMock(return_value=True)
        mock_all = MagicMock(
            return_value={"retcode": True, "stdout": "salt", "stderr": "error"}
        )
        with patch.dict(
            pyenv.__salt__,
            {
                "cmd.has_exec": mock_cmd,
                "config.option": mock_opt,
                "cmd.run_all": mock_all,
            },
        ):
            self.assertListEqual(pyenv.list_(), [])

    # 'rehash' function tests: 1

    def test_rehash(self):
        """
        Test if it run pyenv rehash to update the installed shims.
        """
        mock_opt = MagicMock(return_value="salt stack")
        mock_cmd = MagicMock(return_value=True)
        mock_all = MagicMock(
            return_value={"retcode": True, "stdout": "salt", "stderr": "error"}
        )
        with patch.dict(
            pyenv.__salt__,
            {
                "cmd.has_exec": mock_cmd,
                "config.option": mock_opt,
                "cmd.run_all": mock_all,
            },
        ):
            self.assertTrue(pyenv.rehash())

    # 'do' function tests: 1

    def test_do(self):
        """
        Test if it execute a python command with pyenv's
        shims from the user or the system.
        """
        mock_opt = MagicMock(return_value="salt stack")
        mock_cmd = MagicMock(return_value=True)
        mock_all = MagicMock(
            return_value={"retcode": True, "stdout": "salt", "stderr": "error"}
        )
        with patch.dict(
            pyenv.__salt__,
            {
                "cmd.has_exec": mock_cmd,
                "config.option": mock_opt,
                "cmd.run_all": mock_all,
            },
        ):
            self.assertFalse(pyenv.do("gem list bundler"))

        mock_all = MagicMock(
            return_value={"retcode": 0, "stdout": "salt", "stderr": "error"}
        )
        with patch.dict(
            pyenv.__salt__,
            {
                "config.option": mock_opt,
                "cmd.has_exec": mock_cmd,
                "cmd.run_all": mock_all,
            },
        ):
            self.assertEqual(pyenv.do("gem list bundler"), "salt")

    # 'do_with_python' function tests: 1

    def test_do_with_python(self):
        """
        Test if it execute a python command with pyenv's
        shims using a specific python version.
        """
        mock_opt = MagicMock(return_value="salt stack")
        mock_cmd = MagicMock(return_value=True)
        mock_all = MagicMock(
            return_value={"retcode": True, "stdout": "salt", "stderr": "error"}
        )
        with patch.dict(
            pyenv.__salt__,
            {
                "cmd.has_exec": mock_cmd,
                "config.option": mock_opt,
                "cmd.run_all": mock_all,
            },
        ):
            self.assertFalse(pyenv.do_with_python("2.0.0-p0", "gem list bundler"))



"""
Manage ceph with salt.

.. versionadded:: 2016.11.0
"""

import logging

import salt.utils.json
from salt.exceptions import CommandExecutionError, CommandNotFoundError

log = logging.getLogger(__name__)


def _unchanged(name, msg):
    """
    Utility function: Return structure unchanged
    """
    return {"name": name, "result": True, "comment": msg, "changes": {}}


def _test(name, msg):
    """
    Utility function: Return structure test
    """
    return {"name": name, "result": None, "comment": msg, "changes": {}}


def _error(name, msg):
    """
    Utility function: Return structure error
    """
    return {"name": name, "result": False, "comment": msg, "changes": {}}


def _changed(name, msg, **changes):
    """
    Utility function: Return structure changed
    """
    return {"name": name, "result": True, "comment": msg, "changes": changes}


def _ordereddict2dict(input_ordered_dict):
    """
    Convert ordered dictionary to a dictionary
    """
    return salt.utils.json.loads(salt.utils.json.dumps(input_ordered_dict))


def quorum(name, **kwargs):
    """
    Quorum state

    This state checks the mon daemons are in quorum. It does not alter the
    cluster but can be used in formula as a dependency for many cluster
    operations.

    Example usage in sls file:

    .. code-block:: yaml

        quorum:
          sesceph.quorum:
            - require:
              - sesceph: mon_running
    """
    parameters = _ordereddict2dict(kwargs)
    if parameters is None:
        return _error(name, "Invalid parameters:%s")

    if __opts__["test"]:
        return _test(name, "cluster quorum")
    try:
        cluster_quorum = __salt__["ceph.cluster_quorum"](**parameters)
    except (CommandExecutionError, CommandNotFoundError) as err:
        return _error(name, err.strerror)
    if cluster_quorum:
        return _unchanged(name, "cluster is quorum")
    return _error(name, "cluster is not quorum")



"""
Tests for salt.utils.json
"""

import textwrap

import salt.utils.files
import salt.utils.json
import salt.utils.platform
import salt.utils.stringutils
from tests.support.helpers import with_tempfile
from tests.support.unit import LOREM_IPSUM, TestCase


class JSONTestCase(TestCase):
    data = {
        "спам": "яйца",
        "list": [1, 2, "three"],
        "dict": {"subdict": {"спам": "яйца"}},
        "True": False,
        "float": 1.5,
        "None": None,
    }

    serialized = salt.utils.stringutils.to_str(
        '{"None": null, "True": false, "dict": {"subdict": {"спам": "яйца"}}, "float":'
        ' 1.5, "list": [1, 2, "three"], "спам": "яйца"}'
    )

    serialized_indent4 = salt.utils.stringutils.to_str(
        textwrap.dedent(
            """\
        {
            "None": null,
            "True": false,
            "dict": {
                "subdict": {
                    "спам": "яйца"
                }
            },
            "float": 1.5,
            "list": [
                1,
                2,
                "three"
            ],
            "спам": "яйца"
        }"""
        )
    )

    def test_find_json(self):
        test_sample_json = """
                            {
                                "glossary": {
                                    "title": "example glossary",
                                    "GlossDiv": {
                                        "title": "S",
                                        "GlossList": {
                                            "GlossEntry": {
                                                "ID": "SGML",
                                                "SortAs": "SGML",
                                                "GlossTerm": "Standard Generalized Markup Language",
                                                "Acronym": "SGML",
                                                "Abbrev": "ISO 8879:1986",
                                                "GlossDef": {
                                                    "para": "A meta-markup language, used to create markup languages such as DocBook.",
                                                    "GlossSeeAlso": ["GML", "XML"]
                                                },
                                                "GlossSee": "markup"
                                            }
                                        }
                                    }
                                }
                            }
                            """
        expected_ret = {
            "glossary": {
                "GlossDiv": {
                    "GlossList": {
                        "GlossEntry": {
                            "GlossDef": {
                                "GlossSeeAlso": ["GML", "XML"],
                                "para": (
                                    "A meta-markup language, used to create markup"
                                    " languages such as DocBook."
                                ),
                            },
                            "GlossSee": "markup",
                            "Acronym": "SGML",
                            "GlossTerm": "Standard Generalized Markup Language",
                            "SortAs": "SGML",
                            "Abbrev": "ISO 8879:1986",
                            "ID": "SGML",
                        }
                    },
                    "title": "S",
                },
                "title": "example glossary",
            }
        }

        # First test the valid JSON
        ret = salt.utils.json.find_json(test_sample_json)
        self.assertDictEqual(ret, expected_ret)

        # Now pre-pend some garbage and re-test
        garbage_prepend_json = "{}{}".format(LOREM_IPSUM, test_sample_json)
        ret = salt.utils.json.find_json(garbage_prepend_json)
        self.assertDictEqual(ret, expected_ret)

        # Test to see if a ValueError is raised if no JSON is passed in
        self.assertRaises(ValueError, salt.utils.json.find_json, LOREM_IPSUM)

    def test_dumps_loads(self):
        """
        Test dumping to and loading from a string
        """
        # Dump with no indentation
        ret = salt.utils.json.dumps(self.data, sort_keys=True)
        # Make sure the result is as expected
        self.assertEqual(ret, self.serialized)
        # Loading it should be equal to the original data
        self.assertEqual(salt.utils.json.loads(ret), self.data)

        # Dump with 4 spaces of indentation
        ret = salt.utils.json.dumps(self.data, sort_keys=True, indent=4)
        # Make sure the result is as expected. Note that in Python 2, dumping
        # results in trailing whitespace on lines ending in a comma. So, for a
        # proper comparison, we will have to run rstrip on each line of the
        # return and then stitch it back together.
        ret = "\n".join([x.rstrip() for x in ret.splitlines()])
        self.assertEqual(ret, self.serialized_indent4)
        # Loading it should be equal to the original data
        self.assertEqual(salt.utils.json.loads(ret), self.data)

    @with_tempfile()
    def test_dump_load(self, json_out):
        """
        Test dumping to and loading from a file handle
        """
        with salt.utils.files.fopen(json_out, "wb") as fp_:
            fp_.write(salt.utils.stringutils.to_bytes(salt.utils.json.dumps(self.data)))
        with salt.utils.files.fopen(json_out, "rb") as fp_:
            ret = salt.utils.json.loads(salt.utils.stringutils.to_unicode(fp_.read()))
            # Loading should be equal to the original data
            self.assertEqual(ret, self.data)



"""
tests.integration.modules.pip
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""


import os
import pprint
import re
import shutil
import sys
import tempfile

import pytest
import salt.utils.files
import salt.utils.path
import salt.utils.platform
from salt.modules.virtualenv_mod import KNOWN_BINARY_NAMES
from tests.support.case import ModuleCase
from tests.support.helpers import patched_environ
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import skipIf


@skipIf(
    salt.utils.path.which_bin(KNOWN_BINARY_NAMES) is None, "virtualenv not installed"
)
@pytest.mark.windows_whitelisted
class PipModuleTest(ModuleCase):
    def setUp(self):
        super().setUp()
        self.venv_test_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        # Remove the venv test directory
        self.addCleanup(shutil.rmtree, self.venv_test_dir, ignore_errors=True)
        self.venv_dir = os.path.join(self.venv_test_dir, "venv")
        self.pip_temp = os.path.join(self.venv_test_dir, ".pip-temp")
        if not os.path.isdir(self.pip_temp):
            os.makedirs(self.pip_temp)
        self.patched_environ = patched_environ(
            PIP_SOURCE_DIR="",
            PIP_BUILD_DIR="",
            __cleanup__=[k for k in os.environ if k.startswith("PIP_")],
        )
        self.patched_environ.__enter__()
        self.addCleanup(self.patched_environ.__exit__)

    def _create_virtualenv(self, path):
        """
        The reason why the virtualenv creation is proxied by this function is mostly
        because under windows, we can't seem to properly create a virtualenv off of
        another virtualenv(we can on linux) and also because, we really don't want to
        test virtualenv creation off of another virtualenv, we want a virtualenv created
        from the original python.
        Also, one windows, we must also point to the virtualenv binary outside the existing
        virtualenv because it will fail otherwise
        """
        try:
            if salt.utils.platform.is_windows():
                python = os.path.join(sys.real_prefix, os.path.basename(sys.executable))
            else:
                python_binary_names = [
                    "python{}.{}".format(*sys.version_info),
                    "python{}".format(*sys.version_info),
                    "python",
                ]
                for binary_name in python_binary_names:
                    python = os.path.join(sys.real_prefix, "bin", binary_name)
                    if os.path.exists(python):
                        break
                else:
                    self.fail(
                        "Couldn't find a python binary name under '{}' matching: {}".format(
                            os.path.join(sys.real_prefix, "bin"), python_binary_names
                        )
                    )
            # We're running off a virtualenv, and we don't want to create a virtualenv off of
            # a virtualenv
            kwargs = {"python": python}
        except AttributeError:
            # We're running off of the system python
            kwargs = {}
        self.run_function("virtualenv.create", [path], **kwargs)

    def _check_download_error(self, ret):
        """
        Checks to see if a download error looks transitory
        """
        return any(w in ret for w in ["URLError", "Download error"])

    def pip_successful_install(
        self,
        target,
        expect=(
            "irc3-plugins-test",
            "pep8",
        ),
    ):
        """
        isolate regex for extracting `successful install` message from pip
        """

        expect = set(expect)
        expect_str = "|".join(expect)

        success = re.search(
            r"^.*Successfully installed\s([^\n]+)(?:Clean.*)?", target, re.M | re.S
        )

        success_for = (
            re.findall(r"({})(?:-(?:[\d\.-]))?".format(expect_str), success.groups()[0])
            if success
            else []
        )

        return expect.issubset(set(success_for))

    @pytest.mark.slow_test
    def test_issue_2087_missing_pip(self):
        # Let's create the testing virtualenv
        self._create_virtualenv(self.venv_dir)

        # Let's remove the pip binary
        pip_bin = os.path.join(self.venv_dir, "bin", "pip")
        site_dir = self.run_function(
            "virtualenv.get_distribution_path", [self.venv_dir, "pip"]
        )
        if salt.utils.platform.is_windows():
            pip_bin = os.path.join(self.venv_dir, "Scripts", "pip.exe")
            site_dir = os.path.join(self.venv_dir, "lib", "site-packages")
        if not os.path.isfile(pip_bin):
            self.skipTest("Failed to find the pip binary to the test virtualenv")
        os.remove(pip_bin)

        # Also remove the pip dir from site-packages
        # This is needed now that we're using python -m pip instead of the
        # pip binary directly. python -m pip will still work even if the
        # pip binary is missing
        shutil.rmtree(os.path.join(site_dir, "pip"))

        # Let's run a pip depending functions
        for func in ("pip.freeze", "pip.list"):
            ret = self.run_function(func, bin_env=self.venv_dir)
            self.assertIn(
                "Command required for '{}' not found: "
                "Could not find a `pip` binary".format(func),
                ret,
            )

    @pytest.mark.slow_test
    def test_requirements_as_list_of_chains__cwd_set__absolute_file_path(self):
        self._create_virtualenv(self.venv_dir)

        # Create a requirements file that depends on another one.

        req1_filename = os.path.join(self.venv_dir, "requirements1.txt")
        req1b_filename = os.path.join(self.venv_dir, "requirements1b.txt")
        req2_filename = os.path.join(self.venv_dir, "requirements2.txt")
        req2b_filename = os.path.join(self.venv_dir, "requirements2b.txt")

        with salt.utils.files.fopen(req1_filename, "w") as f:
            f.write("-r requirements1b.txt\n")
        with salt.utils.files.fopen(req1b_filename, "w") as f:
            f.write("irc3-plugins-test\n")
        with salt.utils.files.fopen(req2_filename, "w") as f:
            f.write("-r requirements2b.txt\n")
        with salt.utils.files.fopen(req2b_filename, "w") as f:
            f.write("pep8\n")

        requirements_list = [req1_filename, req2_filename]

        ret = self.run_function(
            "pip.install",
            requirements=requirements_list,
            bin_env=self.venv_dir,
            cwd=self.venv_dir,
        )
        if not isinstance(ret, dict):
            self.fail(
                "The 'pip.install' command did not return the excepted dictionary."
                " Output:\n{}".format(ret)
            )

        try:
            self.assertEqual(ret["retcode"], 0)
            found = self.pip_successful_install(ret["stdout"])
            self.assertTrue(found)
        except KeyError as exc:
            self.fail(
                "The returned dictionary is missing an expected key. Error: '{}'."
                " Dictionary: {}".format(exc, pprint.pformat(ret))
            )

    @pytest.mark.slow_test
    def test_requirements_as_list_of_chains__cwd_not_set__absolute_file_path(self):
        self._create_virtualenv(self.venv_dir)

        # Create a requirements file that depends on another one.

        req1_filename = os.path.join(self.venv_dir, "requirements1.txt")
        req1b_filename = os.path.join(self.venv_dir, "requirements1b.txt")
        req2_filename = os.path.join(self.venv_dir, "requirements2.txt")
        req2b_filename = os.path.join(self.venv_dir, "requirements2b.txt")

        with salt.utils.files.fopen(req1_filename, "w") as f:
            f.write("-r requirements1b.txt\n")
        with salt.utils.files.fopen(req1b_filename, "w") as f:
            f.write("irc3-plugins-test\n")
        with salt.utils.files.fopen(req2_filename, "w") as f:
            f.write("-r requirements2b.txt\n")
        with salt.utils.files.fopen(req2b_filename, "w") as f:
            f.write("pep8\n")

        requirements_list = [req1_filename, req2_filename]

        ret = self.run_function(
            "pip.install", requirements=requirements_list, bin_env=self.venv_dir
        )

        if not isinstance(ret, dict):
            self.fail(
                "The 'pip.install' command did not return the excepted dictionary."
                " Output:\n{}".format(ret)
            )

        try:
            self.assertEqual(ret["retcode"], 0)
            found = self.pip_successful_install(ret["stdout"])
            self.assertTrue(found)
        except KeyError as exc:
            self.fail(
                "The returned dictionary is missing an expected key. Error: '{}'."
                " Dictionary: {}".format(exc, pprint.pformat(ret))
            )

    @pytest.mark.slow_test
    def test_requirements_as_list__absolute_file_path(self):
        self._create_virtualenv(self.venv_dir)

        req1_filename = os.path.join(self.venv_dir, "requirements.txt")
        req2_filename = os.path.join(self.venv_dir, "requirements2.txt")

        with salt.utils.files.fopen(req1_filename, "w") as f:
            f.write("irc3-plugins-test\n")
        with salt.utils.files.fopen(req2_filename, "w") as f:
            f.write("pep8\n")

        requirements_list = [req1_filename, req2_filename]

        ret = self.run_function(
            "pip.install", requirements=requirements_list, bin_env=self.venv_dir
        )

        if not isinstance(ret, dict):
            self.fail(
                "The 'pip.install' command did not return the excepted dictionary."
                " Output:\n{}".format(ret)
            )

        try:
            self.assertEqual(ret["retcode"], 0)
            found = self.pip_successful_install(ret["stdout"])
            self.assertTrue(found)
        except KeyError as exc:
            self.fail(
                "The returned dictionary is missing an expected key. Error: '{}'."
                " Dictionary: {}".format(exc, pprint.pformat(ret))
            )

    @pytest.mark.slow_test
    def test_requirements_as_list__non_absolute_file_path(self):
        self._create_virtualenv(self.venv_dir)

        # Create a requirements file that depends on another one.

        req1_filename = "requirements.txt"
        req2_filename = "requirements2.txt"
        req_cwd = self.venv_dir

        req1_filepath = os.path.join(req_cwd, req1_filename)
        req2_filepath = os.path.join(req_cwd, req2_filename)

        with salt.utils.files.fopen(req1_filepath, "w") as f:
            f.write("irc3-plugins-test\n")
        with salt.utils.files.fopen(req2_filepath, "w") as f:
            f.write("pep8\n")

        requirements_list = [req1_filename, req2_filename]

        ret = self.run_function(
            "pip.install",
            requirements=requirements_list,
            bin_env=self.venv_dir,
            cwd=req_cwd,
        )

        if not isinstance(ret, dict):
            self.fail(
                "The 'pip.install' command did not return the excepted dictionary."
                " Output:\n{}".format(ret)
            )

        try:
            self.assertEqual(ret["retcode"], 0)
            found = self.pip_successful_install(ret["stdout"])
            self.assertTrue(found)
        except KeyError as exc:
            self.fail(
                "The returned dictionary is missing an expected key. Error: '{}'."
                " Dictionary: {}".format(exc, pprint.pformat(ret))
            )

    @pytest.mark.slow_test
    def test_chained_requirements__absolute_file_path(self):
        self._create_virtualenv(self.venv_dir)

        # Create a requirements file that depends on another one.

        req1_filename = os.path.join(self.venv_dir, "requirements.txt")
        req2_filename = os.path.join(self.venv_dir, "requirements2.txt")

        with salt.utils.files.fopen(req1_filename, "w") as f:
            f.write("-r requirements2.txt")
        with salt.utils.files.fopen(req2_filename, "w") as f:
            f.write("pep8")

        ret = self.run_function(
            "pip.install", requirements=req1_filename, bin_env=self.venv_dir
        )
        if not isinstance(ret, dict):
            self.fail(
                "The 'pip.install' command did not return the excepted dictionary."
                " Output:\n{}".format(ret)
            )

        try:
            self.assertEqual(ret["retcode"], 0)
            self.assertIn("installed pep8", ret["stdout"])
        except KeyError as exc:
            self.fail(
                "The returned dictionary is missing an expected key. Error: '{}'."
                " Dictionary: {}".format(exc, pprint.pformat(ret))
            )

    @pytest.mark.slow_test
    def test_chained_requirements__non_absolute_file_path(self):
        self._create_virtualenv(self.venv_dir)

        # Create a requirements file that depends on another one.
        req_basepath = self.venv_dir

        req1_filename = "requirements.txt"
        req2_filename = "requirements2.txt"

        req1_file = os.path.join(self.venv_dir, req1_filename)
        req2_file = os.path.join(self.venv_dir, req2_filename)

        with salt.utils.files.fopen(req1_file, "w") as f:
            f.write("-r requirements2.txt")
        with salt.utils.files.fopen(req2_file, "w") as f:
            f.write("pep8")

        ret = self.run_function(
            "pip.install",
            requirements=req1_filename,
            cwd=req_basepath,
            bin_env=self.venv_dir,
        )
        if not isinstance(ret, dict):
            self.fail(
                "The 'pip.install' command did not return the excepted dictionary."
                " Output:\n{}".format(ret)
            )

        try:
            self.assertEqual(ret["retcode"], 0)
            self.assertIn("installed pep8", ret["stdout"])
        except KeyError as exc:
            self.fail(
                "The returned dictionary is missing an expected key. Error: '{}'."
                " Dictionary: {}".format(exc, pprint.pformat(ret))
            )

    @pytest.mark.slow_test
    def test_issue_4805_nested_requirements(self):
        self._create_virtualenv(self.venv_dir)

        # Create a requirements file that depends on another one.
        req1_filename = os.path.join(self.venv_dir, "requirements.txt")
        req2_filename = os.path.join(self.venv_dir, "requirements2.txt")
        with salt.utils.files.fopen(req1_filename, "w") as f:
            f.write("-r requirements2.txt")
        with salt.utils.files.fopen(req2_filename, "w") as f:
            f.write("pep8")

        ret = self.run_function(
            "pip.install",
            requirements=req1_filename,
            bin_env=self.venv_dir,
            timeout=300,
        )

        if not isinstance(ret, dict):
            self.fail(
                "The 'pip.install' command did not return the excepted dictionary."
                " Output:\n{}".format(ret)
            )

        try:
            if self._check_download_error(ret["stdout"]):
                self.skipTest("Test skipped due to pip download error")
            self.assertEqual(ret["retcode"], 0)
            self.assertIn("installed pep8", ret["stdout"])
        except KeyError as exc:
            self.fail(
                "The returned dictionary is missing an expected key. Error: '{}'."
                " Dictionary: {}".format(exc, pprint.pformat(ret))
            )

    @pytest.mark.slow_test
    def test_pip_uninstall(self):
        # Let's create the testing virtualenv
        self._create_virtualenv(self.venv_dir)
        ret = self.run_function("pip.install", ["pep8"], bin_env=self.venv_dir)

        if not isinstance(ret, dict):
            self.fail(
                "The 'pip.install' command did not return the excepted dictionary."
                " Output:\n{}".format(ret)
            )

        try:
            if self._check_download_error(ret["stdout"]):
                self.skipTest("Test skipped due to pip download error")
            self.assertEqual(ret["retcode"], 0)
            self.assertIn("installed pep8", ret["stdout"])
        except KeyError as exc:
            self.fail(
                "The returned dictionary is missing an expected key. Error: '{}'."
                " Dictionary: {}".format(exc, pprint.pformat(ret))
            )
        ret = self.run_function("pip.uninstall", ["pep8"], bin_env=self.venv_dir)

        if not isinstance(ret, dict):
            self.fail(
                "The 'pip.uninstall' command did not return the excepted dictionary."
                " Output:\n{}".format(ret)
            )

        try:
            self.assertEqual(ret["retcode"], 0)
            self.assertIn("uninstalled pep8", ret["stdout"])
        except KeyError as exc:
            self.fail(
                "The returned dictionary is missing an expected key. Error: '{}'."
                " Dictionary: {}".format(exc, pprint.pformat(ret))
            )

    @pytest.mark.slow_test
    def test_pip_install_upgrade(self):
        # Create the testing virtualenv
        self._create_virtualenv(self.venv_dir)
        ret = self.run_function("pip.install", ["pep8==1.3.4"], bin_env=self.venv_dir)

        if not isinstance(ret, dict):
            self.fail(
                "The 'pip.install' command did not return the excepted dictionary."
                " Output:\n{}".format(ret)
            )

        try:
            if self._check_download_error(ret["stdout"]):
                self.skipTest("Test skipped due to pip download error")
            self.assertEqual(ret["retcode"], 0)
            self.assertIn("installed pep8", ret["stdout"])
        except KeyError as exc:
            self.fail(
                "The returned dictionary is missing an expected key. Error: '{}'."
                " Dictionary: {}".format(exc, pprint.pformat(ret))
            )

        ret = self.run_function(
            "pip.install", ["pep8"], bin_env=self.venv_dir, upgrade=True
        )

        if not isinstance(ret, dict):
            self.fail(
                "The 'pip.install' command did not return the excepted dictionary."
                " Output:\n{}".format(ret)
            )

        try:
            if self._check_download_error(ret["stdout"]):
                self.skipTest("Test skipped due to pip download error")
            self.assertEqual(ret["retcode"], 0)
            self.assertIn("installed pep8", ret["stdout"])
        except KeyError as exc:
            self.fail(
                "The returned dictionary is missing an expected key. Error: '{}'."
                " Dictionary: {}".format(exc, pprint.pformat(ret))
            )

        ret = self.run_function("pip.uninstall", ["pep8"], bin_env=self.venv_dir)

        if not isinstance(ret, dict):
            self.fail(
                "The 'pip.uninstall' command did not return the excepted dictionary."
                " Output:\n{}".format(ret)
            )

        try:
            self.assertEqual(ret["retcode"], 0)
            self.assertIn("uninstalled pep8", ret["stdout"])
        except KeyError as exc:
            self.fail(
                "The returned dictionary is missing an expected key. Error: '{}'."
                " Dictionary: {}".format(exc, pprint.pformat(ret))
            )

    @pytest.mark.slow_test
    def test_pip_install_multiple_editables(self):
        editables = [
            "git+https://github.com/jek/blinker.git#egg=Blinker",
            "git+https://github.com/saltstack/salt-testing.git#egg=SaltTesting",
        ]

        # Create the testing virtualenv
        self._create_virtualenv(self.venv_dir)
        ret = self.run_function(
            "pip.install",
            [],
            editable="{}".format(",".join(editables)),
            bin_env=self.venv_dir,
        )

        if not isinstance(ret, dict):
            self.fail(
                "The 'pip.install' command did not return the excepted dictionary."
                " Output:\n{}".format(ret)
            )

        try:
            if self._check_download_error(ret["stdout"]):
                self.skipTest("Test skipped due to pip download error")
            self.assertEqual(ret["retcode"], 0)
            match = re.search(
                "Successfully installed Blinker(.*) SaltTesting(.*)", ret["stdout"]
            )
            assert match is not None
        except KeyError as exc:
            self.fail(
                "The returned dictionary is missing an expected key. Error: '{}'."
                " Dictionary: {}".format(exc, pprint.pformat(ret))
            )

    @pytest.mark.slow_test
    def test_pip_install_multiple_editables_and_pkgs(self):
        editables = [
            "git+https://github.com/jek/blinker.git#egg=Blinker",
            "git+https://github.com/saltstack/salt-testing.git#egg=SaltTesting",
        ]

        # Create the testing virtualenv
        self._create_virtualenv(self.venv_dir)
        ret = self.run_function(
            "pip.install",
            ["pep8"],
            editable="{}".format(",".join(editables)),
            bin_env=self.venv_dir,
        )

        if not isinstance(ret, dict):
            self.fail(
                "The 'pip.install' command did not return the excepted dictionary."
                " Output:\n{}".format(ret)
            )

        try:
            if self._check_download_error(ret["stdout"]):
                self.skipTest("Test skipped due to pip download error")
            self.assertEqual(ret["retcode"], 0)
            for package in ("Blinker", "SaltTesting", "pep8"):
                self.assertRegex(
                    ret["stdout"],
                    r"(?:.*)(Successfully installed)(?:.*)({})(?:.*)".format(package),
                )
        except KeyError as exc:
            self.fail(
                "The returned dictionary is missing an expected key. Error: '{}'."
                " Dictionary: {}".format(exc, pprint.pformat(ret))
            )

    @skipIf(not os.path.isfile("pip3"), "test where pip3 is installed")
    @skipIf(
        salt.utils.platform.is_windows(), "test specific for linux usage of /bin/python"
    )
    def test_system_pip3(self):
        self.run_function(
            "pip.install", pkgs=["lazyimport==0.0.1"], bin_env="/bin/pip3"
        )
        ret1 = self.run_function("cmd.run", "/bin/pip3 freeze | grep lazyimport")
        self.run_function("pip.uninstall", pkgs=["lazyimport"], bin_env="/bin/pip3")
        ret2 = self.run_function("cmd.run", "/bin/pip3 freeze | grep lazyimport")
        assert "lazyimport==0.0.1" in ret1
        assert ret2 == ""



"""
Return salt data to Zabbix

The following Type: "Zabbix trapper" with "Type of information" Text items are required:

.. code-block:: cfg

    Key: salt.trap.info
    Key: salt.trap.warning
    Key: salt.trap.high

To use the Zabbix returner, append '--return zabbix' to the salt command. ex:

.. code-block:: bash

    salt '*' test.ping --return zabbix
"""


import os

# Define the module's virtual name
__virtualname__ = "zabbix"


def __virtual__():
    if zbx():
        return True
    return False, "Zabbix returner: No zabbix_sender and zabbix_agend.conf found."


def zbx():
    if os.path.exists("/usr/local/zabbix/bin/zabbix_sender") and os.path.exists(
        "/usr/local/zabbix/etc/zabbix_agentd.conf"
    ):
        zabbix_sender = "/usr/local/zabbix/bin/zabbix_sender"
        zabbix_config = "/usr/local/zabbix/etc/zabbix_agentd.conf"
        return {"sender": zabbix_sender, "config": zabbix_config}
    elif os.path.exists("/usr/bin/zabbix_sender") and os.path.exists(
        "/etc/zabbix/zabbix_agentd.conf"
    ):
        zabbix_sender = "/usr/bin/zabbix_sender"
        zabbix_config = "/etc/zabbix/zabbix_agentd.conf"
        return {"sender": zabbix_sender, "config": zabbix_config}
    else:
        return False


def zabbix_send(key, output):
    cmd = (
        zbx()["sender"]
        + " -c "
        + zbx()["config"]
        + " -k "
        + key
        + ' -o "'
        + output
        + '"'
    )
    __salt__["cmd.shell"](cmd)


def save_load(jid, load, minions=None):
    """
    Included for API consistency
    """


def returner(ret):
    changes = False
    errors = False
    job_minion_id = ret["id"]

    if type(ret["return"]) is dict:
        for state, item in ret["return"].items():
            if "comment" in item and "name" in item and item["result"] is False:
                errors = True
                zabbix_send(
                    "salt.trap.high",
                    "SALT:\nname: {}\ncomment: {}".format(
                        item["name"], item["comment"]
                    ),
                )
            elif "comment" in item and "name" in item and item["changes"]:
                changes = True
                zabbix_send(
                    "salt.trap.warning",
                    "SALT:\nname: {}\ncomment: {}".format(
                        item["name"], item["comment"]
                    ),
                )

    if not changes and not errors:
        zabbix_send("salt.trap.info", "SALT {} OK".format(job_minion_id))



"""
Validate the boto_iam module
"""


from tests.support.case import ModuleCase
from tests.support.unit import skipIf

try:
    import boto

    NO_BOTO_MODULE = False
except ImportError:
    NO_BOTO_MODULE = True


@skipIf(
    NO_BOTO_MODULE,
    "Please install the boto library before running boto integration tests.",
)
class BotoIAMTest(ModuleCase):
    def setUp(self):
        try:
            boto.connect_iam()
        except boto.exception.NoAuthHandlerFound:
            self.skipTest(
                "Please setup boto AWS credentials before running boto integration"
                " tests."
            )

    def test_get_account_id(self):
        ret = self.run_function("boto_iam.get_account_id")
        # The AWS account ID is a 12-digit number.
        # http://docs.aws.amazon.com/general/latest/gr/acct-identifiers.html
        self.assertRegex(ret, r"^\d{12}$")



# Python libs

import pytest

# Salt libs
import salt.beacons.service as service_beacon
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {service_beacon: {"__context__": {}, "__salt__": {}}}


def test_non_list_config():
    config = {}

    ret = service_beacon.validate(config)

    assert ret == (False, "Configuration for service beacon must be a list.")


def test_empty_config():
    config = [{}]

    ret = service_beacon.validate(config)

    assert ret == (False, "Configuration for service beacon requires services.")


def test_validate_config_services_none():
    config = [{"services": None}]

    ret = service_beacon.validate(config)

    assert ret == (
        False,
        "Services configuration item for service beacon must be a dictionary.",
    )


def test_validate_config_services_list():
    config = [{"services": [{"sshd": {}}]}]

    ret = service_beacon.validate(config)

    assert ret == (
        False,
        "Services configuration item for service beacon must be a dictionary.",
    )


def test_validate_config_services_valid():
    config = [{"services": {"sshd": {}}}]

    ret = service_beacon.validate(config)

    assert ret == (True, "Valid beacon configuration")


def test_service_running():
    with patch.dict(
        service_beacon.__salt__, {"service.status": MagicMock(return_value=True)}
    ):
        config = [{"services": {"salt-master": {}}}]

        ret = service_beacon.validate(config)

        assert ret == (True, "Valid beacon configuration")

        ret = service_beacon.beacon(config)
        assert ret == [
            {
                "service_name": "salt-master",
                "tag": "salt-master",
                "salt-master": {"running": True},
            }
        ]


def test_service_not_running():
    with patch.dict(
        service_beacon.__salt__, {"service.status": MagicMock(return_value=False)}
    ):
        config = [{"services": {"salt-master": {}}}]

        ret = service_beacon.validate(config)

        assert ret == (True, "Valid beacon configuration")

        ret = service_beacon.beacon(config)
        assert ret == [
            {
                "service_name": "salt-master",
                "tag": "salt-master",
                "salt-master": {"running": False},
            }
        ]



"""
Retrieve EC2 instance data for minions for ec2_tags and ec2_tags_list

The minion id must be the AWS instance-id or value in ``tag_match_key``.  For
example set ``tag_match_key`` to ``Name`` to have the minion-id matched against
the tag 'Name'. The tag contents must be unique. The value of
``tag_match_value`` can be 'uqdn' or 'asis'. if 'uqdn', then the domain will be
stripped before comparison.

Additionally, the ``use_grain`` option can be set to ``True``. This allows the
use of an instance-id grain instead of the minion-id. Since this is a potential
security risk, the configuration can be further expanded to include a list of
minions that are trusted to only allow the alternate id of the instances to
specific hosts. There is no glob matching at this time.

.. note::
    If you are using ``use_grain: True`` in the configuration for this external
    pillar module, the minion must have :conf_minion:`metadata_server_grains`
    enabled in the minion config file (see also :py:mod:`here
    <salt.grains.metadata>`).

    It is important to also note that enabling the ``use_grain`` option allows
    the minion to manipulate the pillar data returned, as described above.

The optional ``tag_list_key`` indicates which keys should be added to
``ec2_tags_list`` and be split by ``tag_list_sep`` (by default ``;``). If a tag
key is included in ``tag_list_key`` it is removed from ec2_tags. If a tag does
not exist it is still included as an empty list.


.. note::
    As with any master configuration change, restart the salt-master daemon for
    changes to take effect.

.. code-block:: yaml

    ext_pillar:
      - ec2_pillar:
          tag_match_key: 'Name'
          tag_match_value: 'asis'
          tag_list_key:
            - Role
          tag_list_sep: ';'
          use_grain: True
          minion_ids:
            - trusted-minion-1
            - trusted-minion-2
            - trusted-minion-3

This is a very simple pillar configuration that simply retrieves the instance
data from AWS. Currently the only portion implemented are EC2 tags, which
returns a list of key/value pairs for all of the EC2 tags assigned to the
instance.
"""


import logging
import re

from salt.utils.versions import StrictVersion as _StrictVersion

try:
    import boto.ec2
    import boto.utils
    import boto.exception

    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

# Set up logging
log = logging.getLogger(__name__)

# DEBUG boto is far too verbose
logging.getLogger("boto").setLevel(logging.WARNING)


def __virtual__():
    """
    Check for required version of boto and make this pillar available
    depending on outcome.
    """
    if not HAS_BOTO:
        return False
    boto_version = _StrictVersion(boto.__version__)
    required_boto_version = _StrictVersion("2.8.0")
    if boto_version < required_boto_version:
        log.error(
            "%s: installed boto version %s < %s, can't retrieve instance data",
            __name__,
            boto_version,
            required_boto_version,
        )
        return False
    return True


def _get_instance_info():
    """
    Helper function to return the instance ID and region of the master where
    this pillar is run.
    """
    identity = boto.utils.get_instance_identity()["document"]
    return (identity["instanceId"], identity["region"])


def ext_pillar(
    minion_id,
    pillar,  # pylint: disable=W0613
    use_grain=False,
    minion_ids=None,
    tag_match_key=None,
    tag_match_value="asis",
    tag_list_key=None,
    tag_list_sep=";",
):
    """
    Execute a command and read the output as YAML
    """
    valid_tag_match_value = ["uqdn", "asis"]

    # meta-data:instance-id
    grain_instance_id = __grains__.get("meta-data", {}).get("instance-id", None)
    if not grain_instance_id:
        # dynamic:instance-identity:document:instanceId
        grain_instance_id = (
            __grains__.get("dynamic", {})
            .get("instance-identity", {})
            .get("document", {})
            .get("instance-id", None)
        )
    if (
        grain_instance_id
        and re.search(r"^i-([0-9a-z]{17}|[0-9a-z]{8})$", grain_instance_id) is None
    ):
        log.error(
            "External pillar %s, instance-id '%s' is not valid for '%s'",
            __name__,
            grain_instance_id,
            minion_id,
        )
        grain_instance_id = None  # invalid instance id found, remove it from use.

    # Check AWS Tag restrictions .i.e. letters, spaces, and numbers and + - = . _ : / @
    if tag_match_key and re.match(r"[\w=.:/@-]+$", tag_match_key) is None:
        log.error(
            "External pillar %s, tag_match_key '%s' is not valid ",
            __name__,
            tag_match_key if isinstance(tag_match_key, str) else "non-string",
        )
        return {}

    if tag_match_key and tag_match_value not in valid_tag_match_value:
        log.error(
            "External pillar %s, tag_value '%s' is not valid must be one of %s",
            __name__,
            tag_match_value,
            " ".join(valid_tag_match_value),
        )
        return {}

    if not tag_match_key:
        base_msg = (
            "External pillar %s, querying EC2 tags for minion id '%s' "
            "against instance-id",
            __name__,
            minion_id,
        )
    else:
        base_msg = (
            "External pillar %s, querying EC2 tags for minion id '%s' "
            "against instance-id or '%s' against '%s'",
            __name__,
            minion_id,
            tag_match_key,
            tag_match_value,
        )

    log.debug(base_msg)
    find_filter = None
    find_id = None

    if re.search(r"^i-([0-9a-z]{17}|[0-9a-z]{8})$", minion_id) is not None:
        find_filter = None
        find_id = minion_id
    elif tag_match_key:
        if tag_match_value == "uqdn":
            find_filter = {"tag:{}".format(tag_match_key): minion_id.split(".", 1)[0]}
        else:
            find_filter = {"tag:{}".format(tag_match_key): minion_id}
        if grain_instance_id:
            # we have an untrusted grain_instance_id, use it to narrow the search
            # even more. Combination will be unique even if uqdn is set.
            find_filter.update({"instance-id": grain_instance_id})
        # Add this if running state is not dependent on EC2Config
        # find_filter.update('instance-state-name': 'running')

    # no minion-id is instance-id and no suitable filter, try use_grain if enabled
    if not find_filter and not find_id and use_grain:
        if not grain_instance_id:
            log.debug(
                "Minion-id is not in AWS instance-id formation, and there "
                "is no instance-id grain for minion %s",
                minion_id,
            )
            return {}
        if minion_ids is not None and minion_id not in minion_ids:
            log.debug(
                "Minion-id is not in AWS instance ID format, and minion_ids "
                "is set in the ec2_pillar configuration, but minion %s is "
                "not in the list of allowed minions %s",
                minion_id,
                minion_ids,
            )
            return {}
        find_id = grain_instance_id

    if not (find_filter or find_id):
        log.debug(
            "External pillar %s, querying EC2 tags for minion id '%s' against "
            "instance-id or '%s' against '%s' noughthing to match against",
            __name__,
            minion_id,
            tag_match_key,
            tag_match_value,
        )
        return {}

    myself = boto.utils.get_instance_metadata(timeout=0.1, num_retries=1)
    if len(myself.keys()) < 1:
        log.info("%s: salt master not an EC2 instance, skipping", __name__)
        return {}

    # Get the Master's instance info, primarily the region
    (_, region) = _get_instance_info()

    # If the Minion's region is available, use it instead
    if use_grain:
        region = __grains__.get("ec2", {}).get("region", region)

    try:
        conn = boto.ec2.connect_to_region(region)
    except boto.exception.AWSConnectionError as exc:
        log.error("%s: invalid AWS credentials, %s", __name__, exc)
        return {}

    if conn is None:
        log.error("%s: Could not connect to region %s", __name__, region)
        return {}

    try:
        if find_id:
            instance_data = conn.get_only_instances(
                instance_ids=[find_id], dry_run=False
            )
        else:
            # filters and max_results can not be used togther.
            instance_data = conn.get_only_instances(filters=find_filter, dry_run=False)

    except boto.exception.EC2ResponseError as exc:
        log.error("%s failed with '%s'", base_msg, exc)
        return {}

    if not instance_data:
        log.debug(
            "%s no match using '%s'", base_msg, find_id if find_id else find_filter
        )
        return {}

    # Find a active instance, i.e. ignore terminated and stopped instances
    active_inst = []
    for idx, inst_data in enumerate(instance_data):
        if inst_data.state not in ["terminated", "stopped"]:
            active_inst.append(idx)

    valid_inst = len(active_inst)
    if not valid_inst:
        log.debug(
            "%s match found but not active '%s'",
            base_msg,
            find_id if find_id else find_filter,
        )
        return {}

    if valid_inst > 1:
        log.error(
            "%s multiple matches, ignored, using '%s'",
            base_msg,
            find_id if find_id else find_filter,
        )
        return {}

    instance = instance_data[active_inst[0]]
    if instance.tags:
        ec2_tags = instance.tags
        ec2_tags_list = {}
        log.debug(
            "External pillar %s, for minion id '%s', tags: %s",
            __name__,
            minion_id,
            instance.tags,
        )
        if tag_list_key and isinstance(tag_list_key, list):
            for item in tag_list_key:
                if item in ec2_tags:
                    ec2_tags_list[item] = ec2_tags[item].split(tag_list_sep)
                    del ec2_tags[item]  # make sure its only in ec2_tags_list
                else:
                    ec2_tags_list[item] = []  # always return a result

        return {"ec2_tags": ec2_tags, "ec2_tags_list": ec2_tags_list}
    return {}



import pytest
import salt.utils.minions
import salt.utils.network
from tests.support.mock import patch


def test_connected_ids():
    """
    test ckminion connected_ids when
    local_port_tcp returns 127.0.0.1
    """
    opts = {"publish_port": 4505, "detect_remote_minions": False}
    minion = "minion"
    ip = salt.utils.network.ip_addrs()
    mdata = {"grains": {"ipv4": ip, "ipv6": []}}
    ckminions = salt.utils.minions.CkMinions({"minion_data_cache": True})
    patch_net = patch("salt.utils.network.local_port_tcp", return_value={"127.0.0.1"})
    patch_list = patch("salt.cache.Cache.list", return_value=[minion])
    patch_fetch = patch("salt.cache.Cache.fetch", return_value=mdata)
    with patch.dict(ckminions.opts, opts):
        with patch_net, patch_list, patch_fetch:
            ret = ckminions.connected_ids()
            assert ret == {minion}


def test_connected_ids_remote_minions():
    """
    test ckminion connected_ids when
    detect_remote_minions is set
    """
    opts = {
        "publish_port": 4505,
        "detect_remote_minions": True,
        "remote_minions_port": 22,
        "minion_data_cache": True,
    }
    minion = "minion"
    minion2 = "minion2"
    minion2_ip = "192.168.2.10"
    ip = salt.utils.network.ip_addrs()
    mdata = {"grains": {"ipv4": ip, "ipv6": []}}
    mdata2 = {"grains": {"ipv4": [minion2_ip], "ipv6": []}}
    patch_net = patch("salt.utils.network.local_port_tcp", return_value={"127.0.0.1"})
    patch_remote_net = patch(
        "salt.utils.network.remote_port_tcp", return_value={minion2_ip}
    )
    patch_list = patch("salt.cache.Cache.list", return_value=[minion, minion2])
    patch_fetch = patch("salt.cache.Cache.fetch", side_effect=[mdata, mdata2])
    ckminions = salt.utils.minions.CkMinions(opts)
    with patch_net, patch_list, patch_fetch, patch_remote_net:
        ret = ckminions.connected_ids()
        assert ret == {minion2, minion}


# These validate_tgt tests make the assumption that CkMinions.check_minions is
# correct. In other words, these tests are only worthwhile if check_minions is
# also correct.
def test_validate_tgt_should_return_false_when_no_valid_minions_have_been_found():
    ckminions = salt.utils.minions.CkMinions(opts={})
    with patch(
        "salt.utils.minions.CkMinions.check_minions", autospec=True, return_value={}
    ):
        result = ckminions.validate_tgt("fnord", "fnord", "fnord", minions=[])
        assert result is False


@pytest.mark.parametrize(
    "valid_minions, target_minions",
    [
        (["one", "two", "three"], ["one", "two", "five"]),
        (["one"], ["one", "two"]),
        (["one", "two", "three", "four"], ["five"]),
    ],
)
def test_validate_tgt_should_return_false_when_minions_have_minions_not_in_valid_minions(
    valid_minions, target_minions
):
    ckminions = salt.utils.minions.CkMinions(opts={})
    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        autospec=True,
        return_value={"minions": valid_minions},
    ):
        result = ckminions.validate_tgt(
            "fnord", "fnord", "fnord", minions=target_minions
        )
        assert result is False


@pytest.mark.parametrize(
    "valid_minions, target_minions",
    [
        (["one", "two", "three", "five"], ["one", "two", "five"]),
        (["one"], ["one"]),
        (["one", "two", "three", "four", "five"], ["five"]),
    ],
)
def test_validate_tgt_should_return_true_when_all_minions_are_found_in_valid_minions(
    valid_minions, target_minions
):
    ckminions = salt.utils.minions.CkMinions(opts={})
    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        autospec=True,
        return_value={"minions": valid_minions},
    ):
        result = ckminions.validate_tgt(
            "fnord", "fnord", "fnord", minions=target_minions
        )
        assert result is True



import logging

from salt.utils.stringutils import camel_to_snake_case

log = logging.getLogger(__name__)

try:
    from testinfra import modules

    TESTINFRA_PRESENT = True
except ImportError:
    TESTINFRA_PRESENT = False

__all__ = []
__virtualname__ = "testinfra"


def __virtual__():
    if TESTINFRA_PRESENT:
        return __virtualname__
    return False, "The Testinfra package is not available"


def _wrap_module_function(func_name):
    def _module_function_wrapper(name, **methods):
        func = __salt__[func_name]
        result, passes, fails = func(name, **methods)
        comment = "\n".join(passes + fails)
        if __opts__["test"] and not result:
            result = None
        return {"name": name, "comment": comment, "result": result, "changes": {}}

    return _module_function_wrapper


def _generate_functions():
    try:
        modules_ = [camel_to_snake_case(module_) for module_ in modules.__all__]
    except AttributeError:
        modules_ = [module_ for module_ in modules.modules]

    for module_name in modules_:
        func_name = "testinfra.{}".format(module_name)
        __all__.append(module_name)
        log.debug(
            "Generating state for module %s as function %s", module_name, func_name
        )
        globals()[module_name] = _wrap_module_function(func_name)


if TESTINFRA_PRESENT:
    _generate_functions()
